
/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

% File: conversions.pl
% Author: Temur Kutsia
% Date: 27.11.2005
% The module for translating input syntax into the internal representation
  and back.
  
Copyright (c) 2005 Temur Kutsia

*************************************************************************/

:-module(conversions,[
        term_matching_problems_to_symbol_list_matching_problems/2,
        symbol_list_substitutions_to_term_substitutions/2,
        remove_bindings_for_temporary_variables/2,
        op(120,yfy,~>),op(110,yf,*)]).


                 /*******************************************
                 *    From input format to internal form    *
                 *******************************************/
                 
% term_to_list transforms a term into its list representation.
% Example: the term f_(a,g__(x__,h(b)),y_) will be translated into
% [f_, a, [g__, x__, [h, b]], y_]

term_to_list(Var,Term):-var(Var),
        term_to_atom(Var,Term),
        !.
term_to_list(seq([]),_):-
        write('Error. seq([]) is not allowed to occur in matching equations.'),
        nl,
        abort.
term_to_list(Ctx_or_seq_or_fun_or_ind_var_or_constant,X):-
        functor(Ctx_or_seq_or_fun_or_ind_var_or_constant,_,0),
        !,
        transform_var_or_constant(Ctx_or_seq_or_fun_or_ind_var_or_constant,X).
term_to_list(Term,List):-Term=..PseudoList,
        pseudo_list_to_list(PseudoList,List).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Transforms variables into terms of the form ctx_or_seq_var(Var) or
 fun_or_ind_var(Var). Leaves constants unchanged.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

        
transform_var_or_constant(Fun_Or_Ind_Var,fun_or_ind_var(Fun_Or_Ind_Var)):-
        fun_or_ind_var(Fun_Or_Ind_Var),
        !.
transform_var_or_constant(Ctx_Or_Seq_Var,ctx_or_seq_var(Ctx_Or_Seq_Var)):-
        ctx_or_seq_var(Ctx_Or_Seq_Var),
        !.
transform_var_or_constant(Term,Term):-functor(Term,_,0),
        !.

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Checks for nonanonymous context, sequence, function and individual
 variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ctx_or_seq_var([_|_]):-!,fail.
ctx_or_seq_var(X):-name(X,N), append(_,[_,95,95],N),!.

fun_or_ind_var([_|_]):-!,fail.
fun_or_ind_var(X):-name(X,N), append(_,[C,95],N),!,C\=95.


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 pseudo_list_to_list takes a list of symbols and terms and returns
 a list of symbols and list representations of terms.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

pseudo_list_to_list([],[]).
pseudo_list_to_list([Term|OtherTerms],[TermList|OtherTermsList]):-
       term_to_list(Term,TermList),
       pseudo_list_to_list(OtherTerms,OtherTermsList).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 term_matching_problems_to_symbol_list_matching_problems transforms
 terms in matching problems into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

term_matching_problems_to_symbol_list_matching_problems([],[]).
term_matching_problems_to_symbol_list_matching_problems([S << T|Rest],
                                                        [SL << TL|RestL]):-
        term_to_list(S,SL),
        term_to_list(T,TL),
        term_matching_problems_to_symbol_list_matching_problems(Rest,RestL).

term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp,
                                                          IntegerFlag]|Rest],
                                                        [[ctx_or_seq_var(Var),
                                                          RegexpL,
                                                          IntegerFlag]|RestL]):-
        regexp_to_list(Regexp,RegexpL),
        term_matching_problems_to_symbol_list_matching_problems(Rest,RestL).
                         
term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp]|Rest],L):-
        term_matching_problems_to_symbol_list_matching_problems([[Var,Regexp,
                                                                  0]|Rest],
                                                                L).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 regexp_to_list transforms terms in regular expressions
 into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

regexp_to_list(R*,star(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list((R1|R2),or(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list((R1,R2),conc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(star(R),star(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list(or(R1,R2),or(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(conc(R1,R2),conc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(seq([]),seq([])):-!.
regexp_to_list(Term,TermList):-
        term_to_list(Term,TermList).
                       

                 /*******************************************
                 *  From internal form to the input format  *
                 *******************************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 symbol_list_substitutions_to_term_substitutions transforms
 symbol list representations of terms in substitutions
 into their input format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

symbol_list_substitutions_to_term_substitutions([],[]).
symbol_list_substitutions_to_term_substitutions([X~>TL |Rest1],
                                                [Y~> T|Rest2]):-
        symbol_list_to_term(X,Y),
        symbol_list_to_term(TL,T),
        symbol_list_substitutions_to_term_substitutions(Rest1,Rest2).


symbol_list_to_term(ctx_or_seq_var(Var),Var):-!.
symbol_list_to_term(fun_or_ind_var(Var),Var):-!.
symbol_list_to_term(seq([]),seq([])):-!.
symbol_list_to_term(seq([X1]),X2):-!,
        symbol_list_to_term(X1,X2).
symbol_list_to_term(seq([H1,H2|Tail]),(X,Y)):-!,
        symbol_list_to_term(H1,X),
        symbol_list_to_term(seq([H2|Tail]),Y).
symbol_list_to_term([Head|Tail],Term):-!,
        list_of_symbol_lists_to_symbol_list(Tail,T1),
        Term=..[Head|T1].
symbol_list_to_term(Term,Term).


list_of_symbol_lists_to_symbol_list([],[]).
list_of_symbol_lists_to_symbol_list([X|T],[X|T1]):-
        functor(X,_,0),
        !,
        list_of_symbol_lists_to_symbol_list(T,T1).
list_of_symbol_lists_to_symbol_list([X|T],[X1|T1]):-
        symbol_list_to_term(X,X1),
        list_of_symbol_lists_to_symbol_list(T,T1).


                 /*******************************************************
                 *   Get substituitions to bind input variables only    *
                 ********************************************************/

               
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_bindings_for_temporary_variables gets a substitution as the input
and removes from it the bindings for all those variables that did not occur
in the input problem but have been generated at run-time.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

remove_bindings_for_temporary_variables([],[]).
remove_bindings_for_temporary_variables([ctx_or_seq_var(Var) ~> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([fun_or_ind_var(Var) ~> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([Var ~> T |Tail],[Var ~> T |Tail1]):-
        remove_bindings_for_temporary_variables(Tail,Tail1).
